@@ -313,7 +313,6 @@ impl std::fmt::Display for Box<dyn TypedOp> {
     }
 }
 
-
 #[derive(Debug, Clone, Hash, PartialEq)]
 pub enum AttrOrInput {
     Attr(Arc<Tensor>),
@@ -328,12 +327,6 @@ impl AttrOrInput {
         }
     }
 
-    fn remove_input(&mut self, ix: usize) {
-        if let AttrOrInput::Input(slot) = self {
-            *slot = *slot - (*slot > ix) as usize;
-        }
-    }
-
     pub fn as_static(&self) -> Option<&Arc<Tensor>> {
         match self {
             AttrOrInput::Attr(t) => Some(t),
@@ -365,5 +358,3 @@ impl<'a> From<&'a Arc<Tensor>> for AttrOrInput {
         AttrOrInput::Attr(t.clone())
     }
 }
-
-
